! msb_model.f90 -- maximum score with bounded regressors
!
! Copyright (C) 2009-2011 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

! This module implements a basic cross-sectional binary choice model.
! The following assumptions are made:
!
! * The coefficient on the first component of X is normalized to 1.
! * The distribution of the first component of X is uniform.
! * The second component of X is discrete.
! * The unobservables are iid and standard Normal distributed.
module msb_model
  use osl_core, wp => osl_wp
  use osl_dist
  use osl_random
  use dataset
  implicit none

  private

  public :: msb_model_t
  public :: msb_model_init
  public :: msb_model_free
  public :: msb_model_generate_data
  public :: msb_model_population_objective
  public :: msb_model_sample_objective

  type :: msb_model_t
     ! True parameters
     real(wp) :: beta0                            ! True parameter

     ! X_1 is Uniform on X1_RANGE
     real(wp), dimension(2) :: x1_range           ! Upper and lower bounds for X_1

     ! Support of X_2
     integer :: n_x2                              ! Number of points in X_2 support
     real(wp), dimension(:), pointer :: x2_sup    ! Support of X_2
     type(osl_discrete_rnd_t) :: x2_dist          ! Discrete X_2 distribution

     ! Simulation draws for X_1
     integer :: nsim                              ! Number of Monte Carlo draws
     type(osl_rng_t) :: rng                       ! Random Number Generator
     real(wp), dimension(:), pointer :: x1_sim    ! Monte Carlo draws
  end type msb_model_t

contains

  ! Initialize a new model and allocate memory.
  subroutine msb_model_init(self, x1_range, x2_sup, beta0, nsim)
    type(msb_model_t), intent(inout) :: self
    real(wp), dimension(2), intent(in) :: x1_range
    real(wp), dimension(:), intent(in) :: x2_sup
    real(wp), intent(in) :: beta0
    integer, intent(in) :: nsim
    integer :: r
    real(wp), dimension(:), allocatable :: px2

    ! Store parameters
    self%beta0 = beta0

    ! Store bounds on support of X_1
    self%x1_range = x1_range

    ! Store the support of X_2
    self%n_x2 = size(x2_sup, 1)
    allocate(self%x2_sup(self%n_x2))
    self%x2_sup = x2_sup

    ! Set up discrete uniform distribution for X_2
    allocate(px2(self%n_x2))
    px2 = 1.0_wp / real(self%n_x2, wp)
    call osl_discrete_rnd_init(self%x2_dist, px2)

    ! Seed the PRNG used for simulation draws
    call osl_rng_seed(self%rng, 9842098)

    ! Simulation draws for X_1
    self%nsim = nsim
    allocate(self%x1_sim(nsim))
    do r = 1, nsim
       self%x1_sim(r) = osl_uniform_rnd(self%rng, x1_range(1), x1_range(2))
    end do
  end subroutine msb_model_init

  ! Destroy an existing model, freeing any allocated memory.
  subroutine msb_model_free(self)
    type(msb_model_t), intent(inout) :: self

    deallocate(self%x1_sim)
    deallocate(self%x2_sup)
    call osl_discrete_rnd_free(self%x2_dist)
  end subroutine msb_model_free

  ! Generate a sample from the model.
  subroutine msb_model_generate_data(self, rng, dat)
    type(msb_model_t), intent(inout) :: self
    type(osl_rng_t), intent(inout) :: rng
    type(dataset_t), intent(inout) :: dat
    integer :: i, ix2
    real(wp) :: xbeta
    real(wp) :: u

    ! Generate data for each individual i
    do i = 1, dat%n
       ! Regressors
       dat%x(1,i) = osl_uniform_rnd(rng, self%x1_range(1), self%x1_range(2))
       ix2 = osl_discrete_rnd(self%x2_dist, rng)
       dat%x(2,i) = self%x2_sup(ix2)

       ! Disturbances
       u = osl_std_normal_rnd(rng)

       ! Determine the choice
       xbeta = dat%x(1,i) + self%beta0 * dat%x(2,i)
       if (xbeta + u > 0.0_wp) then
          dat%y(i) = 1.0_wp
       else
          dat%y(i) = 0.0_wp
       end if
    end do
  end subroutine msb_model_generate_data

  ! Evaluate the population objective function at theta
  subroutine msb_model_population_objective(self, beta, Q)
    type(msb_model_t), intent(in) :: self
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q

    integer :: i, r
    real(wp) :: idx, pr_y
    real(wp), dimension(dim_x) :: x

    ! Calculate expectation of Y * sgn(X'beta) via Monte Carlo integration.
    Q = 0.0_wp
    do i = 1, self%n_x2
       do r = 1, self%nsim
          ! Covariates
          x = (/ self%x1_sim(r), self%x2_sup(i) /)

          ! Evaluate the conditional choice probabilities using theta0
          idx = -x(1) - self%beta0 * x(2)
          pr_y = 1.0_wp - osl_std_normal_cdf(idx)
          pr_y = 2.0_wp * pr_y - 1.0_wp

          ! Evaluate the sgn portion using theta
          idx = x(1) + beta * x(2)

          ! Accumulate the value for this X
          if (idx > 0.0_wp) then
             Q = Q + pr_y
          else
             Q = Q - pr_y
          end if
       end do
    end do

    Q = Q / real(self%n_x2 * self%nsim, wp)
  end subroutine msb_model_population_objective

  ! Evaluate the sample objective function at theta
  subroutine msb_model_sample_objective(self, dat, beta, Q)
    type(msb_model_t), intent(in) :: self
    type(dataset_t), intent(in) :: dat
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q
    integer :: i
    real(wp) :: idx, py

    ! Calculate the sample average of y * sgn(x'beta).
    Q = 0.0_wp
    do i = 1, dat%n
       idx = dat%x(1,i) + beta * dat%x(2,i)
       py = 2.0_wp * dat%y(i) - 1.0_wp
       if (idx > 0.0_wp) then
          Q = Q + py
       else
          Q = Q - py
       end if
    end do
    Q = Q / real(dat%n, wp)
  end subroutine msb_model_sample_objective

end module msb_model
