! osl_core.f90 -- OSL precision, constants, and core procedures
!
! Copyright (C) 2005-2011 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_core
  implicit none

  public

  ! version information
  character(len=*), parameter :: OSL_VERSION = '0.2.4'
  integer, parameter :: OSL_MAJOR_VERSION = 0
  integer, parameter :: OSL_MINOR_VERSION = 2
  integer, parameter :: OSL_MICRO_VERSION = 4

  ! real precision constants
  integer, parameter :: osl_sp = selected_real_kind(6, 37)
  integer, parameter :: osl_dp = selected_real_kind(15, 307)
! integer, parameter :: osl_qp = selected_real_kind(33, 4931)
  integer, parameter :: osl_wp = osl_dp

  ! integer precision constants
  integer, parameter :: osl_int8 = selected_int_kind(2)
  integer, parameter :: osl_int16 = selected_int_kind(4)
  integer, parameter :: osl_int32 = selected_int_kind(9)
  integer, parameter :: osl_int64 = selected_int_kind(18)

  ! internal-use private precision constants
  integer, parameter, private :: sp = osl_sp
  integer, parameter, private :: dp = osl_dp
  integer, parameter, private :: wp = osl_wp

  ! precision constants
  real(sp), parameter :: osl_eps_sp = epsilon(1.0_sp)
  real(dp), parameter :: osl_eps_dp = epsilon(1.0_dp)
  real(wp), parameter :: osl_eps_wp = epsilon(1.0_wp)

  ! mathematical constants
  real(wp), parameter :: &
       OSL_E         = 2.7182818284590452353602874713526625_wp, & ! e
       OSL_LOG2E     = 1.4426950408889634073599246810018921_wp, & ! log_2 e
       OSL_LOG10E    = 0.4342944819032518276511289189166051_wp, & ! log_10 e
       OSL_LN2       = 0.6931471805599453094172321214581766_wp, & ! ln 2
       OSL_LN10      = 2.3025850929940456840179914546843642_wp, & ! ln 10
       OSL_LNPI      = 1.1447298858494001741434273513530587_wp, & ! ln pi
       OSL_PI        = 3.1415926535897932384626433832795029_wp, & ! pi
       OSL_PI_2      = 1.5707963267948966192313216916397514_wp, & ! pi / 2
       OSL_PI_4      = 0.7853981633974483096156608458198757_wp, & ! pi / 4
       OSL_2PI       = 6.2831853071795864769252867665590058_wp, & ! 2 * pi
       OSL_1_PI      = 0.3183098861837906715377675267450287_wp, & ! 1/pi
       OSL_2_PI      = 0.6366197723675813430755350534900574_wp, & ! 2/pi
       OSL_2_SQRTPI  = 1.1283791670955125738961589031215452_wp, & ! 2/sqrt(pi)
       OSL_SQRT1_2   = 0.7071067811865475244008443621048490_wp, & ! 1/sqrt(2)
       OSL_SQRT1_2PI = 0.3989422804014326779399460599343819_wp, & ! 1/sqrt(2*pi)
       OSL_SQRT2     = 1.4142135623730950488016887242096981_wp, & ! sqrt(2)
       OSL_SQRTPI    = 1.7724538509055160272981674833411452_wp, & ! sqrt(pi)
       OSL_SQRT3     = 1.7320508075688772935274463415058724_wp, & ! sqrt(3)
       OSL_SQRT5     = 2.2360679774997896964091736687312762_wp, & ! sqrt(5)
       OSL_SQRT2PI   = 2.5066282746310005024157652848110453_wp, & ! sqrt(2*pi)
       OSL_EULER     = 0.5772156649015328606065120900824024_wp, & ! gamma
       OSL_PHI       = 1.6180339887498948482045868343656381_wp    ! (1+sqrt(5))/2

  ! Default edit descriptors
  character(len=*), parameter :: osl_ed_i   = 'i10'
  character(len=*), parameter :: osl_ed_wp  = 'g17.5'
  integer, parameter, private :: fmt_len    = 80

  ! A context mold for generic interfaces.
  integer, dimension(1), parameter :: osl_ctx = (/ 0 /)

  ! Smallest and largest logical unit numbers
  integer, parameter :: OSL_MIN_UNIT = 10
  integer, parameter :: OSL_MAX_UNIT = 1000

  ! error handling behavior
  integer, parameter :: osl_error_msg_len = 64

  ! status/error codes
  integer, parameter :: OSL_FAILURE  = -1
  integer, parameter :: OSL_SUCCESS  = 0
  integer, parameter :: OSL_EDOM     = 1   ! input domain error
  integer, parameter :: OSL_ERANGE   = 2   ! output range error
  integer, parameter :: OSL_EINVALID = 3   ! invalid argument
  integer, parameter :: OSL_ENOMEM   = 8   ! memory allocation failed
  integer, parameter :: OSL_EMAXITER = 11  ! exceeded maximum iterations
  integer, parameter :: OSL_EMAXEVAL = 12  ! exceeded maximum evaluations
  integer, parameter :: OSL_EUNDRFLW = 15  ! underflow
  integer, parameter :: OSL_EOVRFLW  = 16  ! overflow
  integer, parameter :: OSL_EROUND   = 18  ! roundoff error
  integer, parameter :: OSL_ECONFORM = 19  ! dimensions are non-conforming
  integer, parameter :: OSL_ENOTSQ   = 20  ! matrix is not square
  integer, parameter :: OSL_ESING    = 21  ! singularity detected
  integer, parameter :: OSL_EPOSDEF  = 98  ! matrix not positive definite
  integer, parameter :: OSL_ESYM     = 98  ! matrix not symmetric
  integer, parameter :: OSL_EINIT    = 99  ! object not initialized

  ! Procedures for printing informative messages
  interface osl_print
     module procedure osl_print_str
     module procedure osl_print_swp
     module procedure osl_print_vwp
     module procedure osl_print_mwp
     module procedure osl_print_si
     module procedure osl_print_vi
     module procedure osl_print_mi
  end interface osl_print

  ! Norms
  interface osl_norm_2
     module procedure osl_norm_2_vwp
  end interface osl_norm_2

  interface osl_norm_inf
     module procedure osl_norm_inf_vwp
     module procedure osl_norm_inf_mwp
  end interface osl_norm_inf

  ! Miscellaneous Mathematical Functions
  interface osl_sgn
     module procedure osl_sgn_i
     module procedure osl_sgn_wp
  end interface osl_sgn

  interface osl_clamp
     module procedure osl_clamp_swp
     module procedure osl_clamp_vswp
     module procedure osl_clamp_vvwp
  end interface osl_clamp

  ! IEEE floating point procedures
  interface osl_is_inf
     module procedure osl_is_inf_wp
  end interface osl_is_inf

  ! Swap
  interface osl_swap
     module procedure osl_swap_scalar_wp
     module procedure osl_swap_scalar_i
     module procedure osl_swap_vector_wp
     module procedure osl_swap_vector_i
  end interface osl_swap

  ! Reverse
  interface osl_reverse
     module procedure osl_reverse_vector_wp
     module procedure osl_reverse_vector_i
  end interface osl_reverse

  ! Linear spaced vectors
  interface osl_linspace
     module procedure osl_linspace_swp
     module procedure osl_linspace_vwp
  end interface osl_linspace

contains

  !---------------------------------------------------------------------------!
  ! Version information                                                       !
  !---------------------------------------------------------------------------!

  ! Checks the version of the OSL.  Returns .TRUE. if the
  ! version of the library is at least as new as the given
  ! version and .FALSE. otherwise.
  function osl_check_version(major, minor, micro) result(check)
    integer, intent(in) :: major, minor, micro
    logical :: check
    check = &
         (OSL_MAJOR_VERSION >= major) .or. &
         (OSL_MAJOR_VERSION == major .and. OSL_MINOR_VERSION >= minor) .or. &
         (OSL_MAJOR_VERSION == major .and. OSL_MINOR_VERSION == minor .and. &
          OSL_MICRO_VERSION >= micro)
  end function osl_check_version

  !--------------------------------------------------------------------------!
  ! Basic Input and Output                                                   !
  !--------------------------------------------------------------------------!

  ! Return a free logical unit between OSL_MIN_UNIT and OSL_MAX_UNIT.
  ! If no free unit is found, return -1.
  function osl_get_unit() result(unit)
    integer :: unit
    logical :: used

    do unit = OSL_MIN_UNIT, OSL_MAX_UNIT
       inquire(UNIT=unit, OPENED=used)
       if (.not. used) return
    end do
    unit = -1
  end function osl_get_unit

  !--------------------------------------------------------------------------!
  ! Formatted output                                                         !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! Informative messages               !
  !------------------------------------!

  ! Print MSG followed by a REAL(WP) matrix A.  If MSG is the empty string,
  ! then no blank line will be printed.  If FMT is present, it will be used
  ! to format each element of A.  Otherwise, OSL_ED_WP, the default edit
  ! descriptor for REAL(WP) values, will be used.
  subroutine osl_print_mwp(msg, A, fmt)
    character(len=*), intent(in) :: msg
    real(wp), dimension(:,:), intent(in) :: A
    character(len=*), intent(in), optional :: fmt
    character(len=fmt_len) :: row_fmt
    integer :: i

    if (len(msg) > 0) then
       print '(a)', msg
    end if

    if (present(fmt)) then
       write(row_fmt, '("(",i0,"(",a,"))")') size(A,2), fmt
    else
       write(row_fmt, '("(",i0,"(",a,"))")') size(A,2), osl_ed_wp
    end if

    do i = 1, size(A,1)
       print row_fmt, A(i,:)
    end do
  end subroutine osl_print_mwp

  ! Print MSG followed by an INTEGER matrix A.  If MSG is the empty string,
  ! then no blank line will be printed.  If FMT is present, it will be used
  ! to format each element of A.  Otherwise, OSL_ED_I, the default edit
  ! descriptor for INTEGER values, will be used.
  subroutine osl_print_mi(msg, A, fmt)
    character(len=*), intent(in) :: msg
    integer, dimension(:,:), intent(in) :: A
    character(len=*), intent(in), optional :: fmt
    character(len=fmt_len) :: row_fmt
    integer :: i

    if (len(msg) > 0) then
       print '(a)', msg
    end if

    if (present(fmt)) then
       write(row_fmt, '("(",i0,"(",a,"))")') size(A,2), fmt
    else
       write(row_fmt, '("(",i0,"(",a,"))")') size(A,2), osl_ed_i
    end if

    do i = 1, size(A,1)
       print row_fmt, A(i,:)
    end do
  end subroutine osl_print_mi

  ! Print MSG followed by a REAL(WP) vector VEC.  At most ten elements
  ! will be printed per line.  If MSG is the empty string, then no
  ! blank line will be printed.  If FMT is present, it will be used to
  ! format each element of VEC.  Otherwise, OSL_ED_WP, the default edit
  ! descriptor for REAL(WP) values, will be used.
  subroutine osl_print_vwp(msg, vec, fmt)
    character(len=*), intent(in) :: msg
    real(wp), dimension(:), intent(in) :: vec
    character(len=*), intent(in), optional :: fmt
    character(len=fmt_len) :: fmt_str
    integer :: i, ll, ncols, lines

    if (len(msg) > 0) then
       print '(a)', msg
    end if

    if (present(fmt)) then
       fmt_str = '(10' // fmt // ')'
    else
       fmt_str = '(10' // osl_ed_wp // ')'
    end if

    ncols = size(vec, 1)
    ll = 0
    if (ncols > 10) then
       lines = ncols / 10
       do i = 1, lines
          ll = 10 * (i - 1)
          print fmt_str, vec(1+ll:10+ll)
       end do
       print fmt_str, vec(11+ll:ncols)
    else
       print fmt_str, vec
    end if
  end subroutine osl_print_vwp

  ! Print MSG followed by an INTEGER vector VEC.  At most ten elements
  ! will be printed per line.  If MSG is the empty string, then no
  ! blank line will be printed.  If FMT is present, it will be used to
  ! format each element of VEC.  Otherwise, OSL_ED_I, the default edit
  ! descriptor for INTEGER values, will be used.
  subroutine osl_print_vi(msg, vec, fmt)
    character(len=*), intent(in) :: msg
    integer, dimension(:), intent(in) :: vec
    character(len=*), intent(in), optional :: fmt
    character(len=fmt_len) :: fmt_str
    integer :: i, ll, ncols, lines

    if (len(msg) > 0) then
       print '(a)', msg
    end if

    if (present(fmt)) then
       fmt_str = '(10' // fmt // ')'
    else
       fmt_str = '(10' // osl_ed_i // ')'
    end if

    ncols = size(vec, 1)
    ll = 0
    if (ncols > 10) then
       lines = ncols / 10
       do i = 1, lines
          ll = 10 * (i - 1)
          print fmt_str, vec(1+ll:10+ll)
       end do
       print fmt_str, vec(11+ll:ncols)
    else
       print fmt_str, vec
    end if
  end subroutine osl_print_vi

  ! Print MSG followed by the REAL(WP) scalar X.  If FMT is present,
  ! it will be used to format X.  Otherwise, OSL_ED_WP, the default
  ! edit descriptor for REAL(WP) values, will be used.
  subroutine osl_print_swp(msg, x, fmt)
    character(len=*), intent(in) :: msg
    real(wp), intent(in) :: x
    character(len=*), intent(in), optional :: fmt

    if (present(fmt)) then
       print '(a,' // fmt // ')', msg, x
    else
       print '(a,' // osl_ed_wp // ')', msg, x
    end if
  end subroutine osl_print_swp

  ! Print MSG followed by the INTEGER scalar X.  If FMT is present,
  ! it will be used to format X.  Otherwise, OSL_ED_I, the default
  ! edit descriptor for INTEGER values, will be used.
  subroutine osl_print_si(msg, x, fmt)
    character(len=*), intent(in) :: msg
    integer, intent(in) :: x
    character(len=*), intent(in), optional :: fmt

    if (present(fmt)) then
       print '(a,' // fmt // ')', msg, x
    else
       print '(a,' // osl_ed_i // ')', msg, x
    end if
  end subroutine osl_print_si

  ! Prints a single string.
  subroutine osl_print_str(msg)
    character(len=*), intent(in) :: msg
    print '(a)', msg
  end subroutine osl_print_str

  !------------------------------------!
  ! Headings                           !
  !------------------------------------!

  ! Print a heading consisting of TEXT underlined by equals signs.
  subroutine osl_print_heading(text)
    character(len=*), intent(in) :: text
    print '(/,a,/,a,/)', text, repeat("=", len_trim(text))
  end subroutine osl_print_heading

  ! Print a subheading consisting of TEXT underlined by hyphens.
  subroutine osl_print_subheading(text)
    character(len=*), intent(in) :: text
    print '(/,a,/,a,/)', text, repeat("-", len_trim(text))
  end subroutine osl_print_subheading

  !--------------------------------------------------------------------------!
  ! Error Handling                                                           !
  !--------------------------------------------------------------------------!

  ! Return a string describing the error with number ERRNO.
  function osl_error_msg(errno) result(msg)
    integer, intent(in) :: errno
    character(len=osl_error_msg_len) :: msg

    select case (errno)
    case (OSL_FAILURE)
       msg = 'Generic failure'
    case (OSL_SUCCESS)
       msg = 'Generic success'
    case (OSL_EDOM)
       msg = 'Domain (input) error'
    case (OSL_ERANGE)
       msg = 'Range (output) error'
    case (OSL_EINVALID)
       msg = 'Invalid argument'
    case (OSL_ENOMEM)
       msg = 'Memory allocation failed'
    case (OSL_EMAXITER)
       msg = 'Maximum number of iterations exceeded'
    case (OSL_EMAXEVAL)
       msg = 'Maximum number of evaluations exceeded'
    case (OSL_EUNDRFLW)
       msg = 'Numerical underflow'
    case (OSL_EOVRFLW)
       msg = 'Numerical overflow'
    case (OSL_ECONFORM)
       msg = 'Non-conformant dimensions'
    case (OSL_ENOTSQ)
       msg = 'Matrix is not square'
    case (OSL_EINIT)
       msg = 'Object not initialized'
    case default
       msg = 'Unknown error'
    end select
  end function osl_error_msg

  ! If STATUS is present, simply copy ERRNO to STATUS and return.
  ! Otherwise, print an error message and terminate the program.
  subroutine osl_status(errno, status, reason)
    integer, intent(in) :: errno
    integer, intent(out), optional :: status
    character(len=*), intent(in), optional :: reason

    if (present(status)) then
       status = errno
    else if (errno /= OSL_SUCCESS) then
       if (present(reason)) then
          print "('ERROR: ', a, ' (', i2.2, ')')", reason, errno
       else
          print "('ERROR: ', a, ' (', i2.2, ')')", osl_error_msg(errno), errno
       end if
       stop
    end if
  end subroutine osl_status

  !--------------------------------------------------------------------------!
  ! Norms                                                                    !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! L_2 Norm                           !
  !------------------------------------!

  ! Returns the Euclidean norm of a vector defined as
  ! SQRT(DOT_PRODUCT(x,x)).  This function is based on the BLAS
  ! Fortran 77 implementation of DNRM2 which attempts to prevent
  ! overflow.
  function osl_norm_2_vwp(x) result(norm)
    real(wp), dimension(:), intent(in) :: x
    real(wp) :: norm
    real(wp) :: absxi, scale, ssq
    real(wp), parameter :: ONE = 1.0_wp, ZERO = 0.0_wp
    integer :: i, n

    n = size(x)
    if (n == 1) then
       norm = abs(x(1))
    else
       scale = ZERO
       ssq = ONE
       do i = 1, n
          if (x(i) /= ZERO) THEN
             absxi = abs(x(i))
             if (scale < absxi) then
                ssq = ONE + ssq * (scale / absxi)**2
                scale = absxi
             else
                ssq = ssq + (absxi / scale)**2
             end if
          end if
       end do
       norm = scale * sqrt(ssq)
    end if
  end function osl_norm_2_vwp

  !------------------------------------!
  ! L_\infty Norm                      !
  !------------------------------------!

  ! Calculate the L_\infty norm of a vector.
  function osl_norm_inf_vwp(x) result(norm)
    real(wp), dimension(:), intent(in) :: x
    real(wp) :: norm
    norm = maxval(abs(x))
  end function osl_norm_inf_vwp

  ! Calculate the L_\infty norm of a matrix.
  function osl_norm_inf_mwp(x) result(norm)
    real(wp), dimension(:,:), intent(in) :: x
    real(wp) :: norm
    norm = maxval(abs(x))
  end function osl_norm_inf_mwp

  !--------------------------------------------------------------------------!
  ! Miscellaneous Mathematical Functions                                     !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! The Signum Function                !
  !------------------------------------!

  ! Return 1 if the X is postive, -1 if X is negative, and 0 if X is zero.
  elemental function osl_sgn_i(x) result(sgn)
    integer, intent(in) :: x
    integer :: sgn

    if (x < 0) then
       sgn = -1
    else if (x > 0) then
       sgn = 1
    else
       sgn = 0
    end if
  end function osl_sgn_i

  ! Return 1 if the X is postive, -1 if X is negative, and 0 if X is zero.
  elemental function osl_sgn_wp(x) result(sgn)
    real(wp), intent(in) :: x
    integer :: sgn

    if (x < 0.0_wp) then
       sgn = -1
    else if (x > 0.0_wp) then
       sgn = 1
    else
       sgn = 0
    end if
  end function osl_sgn_wp

  !------------------------------------!
  ! Clamp                              !
  !------------------------------------!

  ! Returns the value of X bounded to be between A and B.
  function osl_clamp_swp(x, a, b) result(y)
    real(wp), intent(in) :: x, a, b
    real(wp) :: y

    if (x < a) then
       y = a
    else if (x < b) then
       y = x
    else
       y = b
    end if
  end function osl_clamp_swp

  ! Returns the vector X with elements bounded to be between A and B.
  function osl_clamp_vswp(x, a, b) result(y)
    real(wp), dimension(:), intent(in) :: x
    real(wp), intent(in) :: a, b
    real(wp), dimension(size(x)) :: y

    where (x < a)
       y = a
    elsewhere (x < b)
       y = x
    elsewhere
       y = b
    end where
  end function osl_clamp_vswp

  ! Returns the vector X with elements bounded to be between A and B.
  function osl_clamp_vvwp(x, a, b) result(y)
    real(wp), dimension(:), intent(in) :: x, a, b
    real(wp), dimension(size(x)) :: y

    where (x < a)
       y = a
    elsewhere (x < b)
       y = x
    elsewhere
       y = b
    end where
  end function osl_clamp_vvwp

  !--------------------------------------------------------------------------!
  ! IEEE Floating Point Arithmetic                                           !
  !--------------------------------------------------------------------------!

  ! Return .TRUE. if the given real variable is +Inf or -Inf and
  ! .FALSE. otherwise.
  elemental function osl_is_inf_wp(x)
    real(wp), intent(in):: x
    logical :: osl_is_inf_wp

    osl_is_inf_wp = (abs(x) > huge(x))
  end function osl_is_inf_wp

  !--------------------------------------------------------------------------!
  ! Swap                                                                     !
  !--------------------------------------------------------------------------!

  ! Swap the values of two REAL(WP) scalars.
  subroutine osl_swap_scalar_wp(a, b)
    real(wp), intent(inout) :: a, b
    real(wp) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_scalar_wp

  ! Swap the values of two INTEGER scalars.
  subroutine osl_swap_scalar_i(a, b)
    integer, intent(inout) :: a, b
    integer :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_scalar_i

  ! Swaps the contents of two REAL(WP) vectors.
  subroutine osl_swap_vector_wp(a, b)
    real(wp), dimension(:), intent(inout) :: a, b
    real(wp), dimension(size(a)) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_vector_wp

  ! Swaps the contents of two INTEGER vectors.
  subroutine osl_swap_vector_i(a, b)
    integer, dimension(:), intent(inout) :: a, b
    integer, dimension(size(a)) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_vector_i

  !--------------------------------------------------------------------------!
  ! Reverse                                                                  !
  !--------------------------------------------------------------------------!

  ! Reverse the elements of a REAL(WP) vector A.
  subroutine osl_reverse_vector_wp(a)
    real(wp), dimension(:), intent(inout) :: a
    real(wp), dimension(size(a, 1)) :: reverse
    integer :: i

    do i = 1, size(a, 1)
       reverse(size(a, 1) - i + 1) = a(i)
    end do
    a = reverse
  end subroutine osl_reverse_vector_wp

  ! Reverse the elements of an INTEGER vector A.
  subroutine osl_reverse_vector_i(a)
    integer, dimension(:), intent(inout) :: a
    integer, dimension(size(a, 1)) :: reverse
    integer :: i

    do i = 1, size(a, 1)
       reverse(size(a, 1) - i + 1) = a(i)
    end do
    a = reverse
  end subroutine osl_reverse_vector_i

  !--------------------------------------------------------------------------!
  ! Vector-vector operations                                                 !
  !--------------------------------------------------------------------------!

  ! Return the outer product of the vectors A and B.
  function osl_outer_product(a, b) result(C)
    real(wp), dimension(:), intent(in) :: a
    real(wp), dimension(:), intent(in) :: b
    real(wp), dimension(size(a), size(b)) :: C
    C = spread(a, DIM=2, NCOPIES=size(b)) * spread(b, DIM=1, NCOPIES=size(a))
  end function osl_outer_product

  !--------------------------------------------------------------------------!
  ! Elementary Arrays                                                        !
  !--------------------------------------------------------------------------!

  ! Return N evenly-spaced scalars from A to B.
  function osl_linspace_swp(a, b, n) result(v)
    real(wp), intent(in) :: a, b
    integer, intent(in) :: n
    real(wp), dimension(n) :: v
    real(wp) :: step
    integer :: i

    step = (b - a) / real(n - 1, wp)
    v(1) = a
    do i = 2, n
       v(i) = v(i-1) + step
    end do
  end function osl_linspace_swp

  ! Return N evenly-spaced vectors from A to B.
  function osl_linspace_vwp(a, b, n) result(m)
    real(wp), dimension(:), intent(in) :: a, b
    integer, intent(in) :: n
    real(wp), dimension(n,size(a)) :: m
    real(wp), dimension(size(a)) :: step
    integer :: i

    step = (b - a) / real(n - 1, wp)
    m(1,:) = a
    do i = 2, n
       m(i,:) = m(i-1,:) + step
    end do
  end function osl_linspace_vwp

end module osl_core
