module utility
  use osl_core, wp => osl_wp
  implicit none

  interface print_vec
     module procedure print_vec_i
     module procedure print_vec_wp
  end interface print_vec

  ! Timer variables
  integer :: count0, count1, count_rate, time
  real(wp) :: time0, time1

contains

  ! Returns .TRUE. if the interval A covers B and .FALSE. otherwise.
  function cover_p(a, b)
    logical, dimension(:), intent(in) :: a, b
    logical :: cover_p
    integer :: n, i

    n = size(a, 1)
    if (size(b, 1) /= n) then
       stop 'Dimension mismatch in COVER_P!'
    end if

    cover_p = .true.
    do i = 1, size(a, 1)
       if (b(i) .and. .not. a(i)) then
          cover_p = .false.
          exit
       end if
    end do
  end function cover_p

  ! Calculates the Hausdorff distance between two intervals A and B.
  function hausdorff(a, b) result(d)
    real(wp), dimension(2), intent(in) :: a
    real(wp), dimension(2), intent(in) :: b
    real(wp) :: d
    d = maxval(abs(a - b))
  end function hausdorff

  ! Calculates the P-th quantile of a _sorted_ vector of real numbers
  ! X using interpolation:
  !
  !     quantile = (1 - \delta) x_i + \delta x_{i+1}
  !
  ! where i is floor(n * p) and \delta is n * p - i.  Thus,
  ! when P is zero, the minimum of the data is returned, when P is
  ! one, the maximum value is returned, and when P is 0.5, the median
  ! value is returned.
  function quantile(x, p)
    real(wp), dimension(:), intent(in) :: x
    real(wp),               intent(in) :: p

    real(wp) :: quantile
    integer :: n, lhs
    real(wp) :: idx, delta

    n = size(x, 1)
    idx = p * n
    lhs = int(idx)
    delta = idx - real(lhs, wp)

    if (n == 0) then
       quantile = 0.0_wp
    else if (lhs == n) then
       quantile = x(lhs)
    else
       quantile = (1.0_wp - delta) * x(lhs) + delta * x(lhs + 1)
    end if
  end function quantile

  ! Determine the lower and upper bounds of the interval, returned in
  ! SET, for which the functional values Y over the grid points
  ! X are larger than KAPPA.
  subroutine get_contour_set(x, y, kappa, ind, bounds)
    real(wp), dimension(:), intent(in) :: x
    real(wp), dimension(size(x)), intent(in) :: y
    real(wp), intent(in) :: kappa
    logical, dimension(size(x)), intent(out) :: ind
    real(wp), dimension(2), intent(out) :: bounds
    integer :: i

    ind = (y >= kappa)
    if (any(ind)) then
        bounds(1) = minval(x, MASK=ind, DIM=1)
        bounds(2) = maxval(x, MASK=ind, DIM=1)
    else
        bounds(1) = x(1)
        bounds(2) = x(size(x))
    end if
  end subroutine get_contour_set

  ! Print REAL(WP) vector with a prefix, postfix, and custom format
  ! and separator.
  subroutine print_vec_i(pre, vec, fmt, sep, post)
    character(len=*),               intent(in) :: pre
    integer,          dimension(:), intent(in) :: vec
    character(len=*),               intent(in) :: fmt
    character(len=*),               intent(in) :: sep
    character(len=*),               intent(in) :: post
    integer :: i, n

    n = size(vec)
    write (*, '(a)', advance='no') pre
    do i = 1, n - 1
       write (*, '(' // fmt // ')', advance='no') vec(i)
       write (*, '(a)', advance='no') sep
    end do
    write (*, '(' // fmt // ')', advance='no') vec(n)
    write (*, '(a)') post
  end subroutine print_vec_i

  ! Print REAL(WP) vector with a prefix, postfix, and custom format
  ! and separator.
  subroutine print_vec_wp(pre, vec, fmt, sep, post)
    character(len=*),               intent(in) :: pre
    real(wp),         dimension(:), intent(in) :: vec
    character(len=*),               intent(in) :: fmt
    character(len=*),               intent(in) :: sep
    character(len=*),               intent(in) :: post
    integer :: i, n

    n = size(vec)
    write (*, '(a)', advance='no') pre
    do i = 1, n - 1
       write (*, '(' // fmt // ')', advance='no') vec(i)
       write (*, '(a)', advance='no') sep
    end do
    write (*, '(' // fmt // ')', advance='no') vec(n)
    write (*, '(a)') post
  end subroutine print_vec_wp

  ! Start timer
  subroutine tic()
    call system_clock(count0)
    call cpu_time(time0)
  end subroutine tic

  ! Stop timer and report elapsed wall clock and CPU time
  subroutine toc()
    call system_clock(count1, count_rate)
    call cpu_time(time1)
    time = int((count1 - count0) / real(count_rate))
    print '(a,i0,a,i0,a)', '% Elapsed time: ', &
         time, ' sec. (', int(time1 - time0), ' sec. cpu time)'
    print '(a)',    '%'
  end subroutine toc

end module utility
